#include <stdio.h>
#include <gtk/gtk.h>
#include <math.h>
#include <errno.h>
#include <stdlib.h>
#include <CL/cl.h>
#include <epoxy/gl.h>

struct canvas_buffer {
		unsigned int *mandelbrot_buffer;
		unsigned int iterations;
		unsigned int height;
		unsigned int width;
		double center_x;
		double center_y;
		double x_span;
		double y_span;
};

static gboolean on_main_window_close(GtkWidget *window, gpointer data)
{
	gtk_main_quit();
	return FALSE;
}

#define check_error(error) \
	do { \
	if (error != CL_SUCCESS) { \
		printf("OpenCL call failed with error %d. %s:%d\n", (int)error, __func__, __LINE__); \
		exit(error); \
	} } while(0)


void load_kernel_from_file(char *file, char **src)
{
	long fsize;
	FILE *f = fopen(file, "rb");

	fseek(f, 0, SEEK_END);
	fsize = ftell(f);
	fseek(f, 0, SEEK_SET);  //same as rewind(f);
	*src = (char *)malloc(fsize + 1);
	fread(*src, fsize, 1, f);
	fclose(f);
	(*src)[fsize] = 0;
}

cl_program create_program(char *src, cl_context context)
{
	cl_program program;
	size_t lengths[1];
	cl_int error = 0;

	lengths[0] = strlen(src);
	printf("strlen: %u\n", (unsigned int)lengths[0]);
	program = clCreateProgramWithSource(context, 1, (const char **)&src,
					    lengths, &error);
	check_error(error);

	return program;
}

static int calculate_mandelbrot(struct canvas_buffer *buff)
{

	int i;
	cl_int error;
	cl_int ret;
	cl_platform_id *platform_ids = NULL;
	cl_device_id *device_ids = NULL;
	cl_uint platform_id_count = 0;
	cl_uint device_id_count = 0;
	cl_context context;
	cl_program program;
	cl_kernel kernel;
	cl_command_queue queue;
	cl_mem dat_buff;
	char *source_code = NULL;
	char temp_buff[100];
	size_t data_size;
	size_t work_size = 64;
	unsigned int *data_array;

	data_size = buff->width*buff->height;

	if (!buff->mandelbrot_buffer)
		buff->mandelbrot_buffer = (unsigned int *)calloc(data_size, sizeof(buff->mandelbrot_buffer[0]));

	data_array = buff->mandelbrot_buffer;
	if (!data_array)
		return -ENOMEM;


	clGetPlatformIDs(0, NULL, &platform_id_count);
	if (platform_id_count == 0)
		return -ENOMEM;

	platform_ids = (cl_platform_id *)malloc(platform_id_count *
						sizeof(cl_platform_id));
	if (platform_ids == NULL)
		return -ENOMEM;

	clGetPlatformIDs(platform_id_count, platform_ids, NULL);

	printf("Platforms available: %u\n", (unsigned int)platform_id_count);

	for (i = 0; i < platform_id_count; i++) {
		temp_buff[0] = '\0';
		clGetPlatformInfo(platform_ids[i], CL_PLATFORM_NAME, sizeof(temp_buff), temp_buff, NULL);
		printf("Platform ID %d: %s\n", i, temp_buff);
	}

	clGetDeviceIDs(platform_ids[0], CL_DEVICE_TYPE_ALL, 0, NULL,
			&device_id_count);
	printf("Device in Platform 0 count: %u\n",
	       (unsigned int)device_id_count);

	if (device_id_count == 0)
		return -1;

	device_ids = (cl_device_id *) malloc(sizeof(cl_device_id)
					     *device_id_count);

	clGetDeviceIDs(platform_ids[0], CL_DEVICE_TYPE_ALL, device_id_count,
			device_ids, NULL);

	for (i = 0; i < device_id_count; i++) {
		temp_buff[0] = '\0';
		clGetDeviceInfo(device_ids[i], CL_DEVICE_NAME, sizeof(temp_buff), temp_buff, NULL);
		printf("Device ID %d: %s\n", i, temp_buff);
	}


	context = clCreateContext(NULL, 1, &device_ids[0], NULL, NULL, &error);
	check_error(error);

	queue = clCreateCommandQueue(context, device_ids[0], 0, &error);

	dat_buff = clCreateBuffer(context, CL_MEM_WRITE_ONLY |
					CL_MEM_COPY_HOST_PTR,
					sizeof(data_array[0])*data_size, data_array, &error);
	check_error(error);

	load_kernel_from_file("cl_kernels/mandelbrot.cl", &source_code);
	program = create_program(source_code, context);

	/* Compile the source code */
	check_error(clBuildProgram(program, device_id_count,
					   device_ids, NULL, NULL, NULL));
	kernel = clCreateKernel(program, "mandelbrot", &error);


	/* Setting the arguments */
	clSetKernelArg(kernel, 0, sizeof(cl_mem), (void *)&dat_buff);
	clSetKernelArg(kernel, 1, sizeof(unsigned int), (void *)&buff->iterations);
	clSetKernelArg(kernel, 2, sizeof(double), (void *)&buff->center_x);
	clSetKernelArg(kernel, 3, sizeof(double), (void *)&buff->center_y);
	clSetKernelArg(kernel, 4, sizeof(double), (void *)&buff->x_span);
	clSetKernelArg(kernel, 5, sizeof(double), (void *)&buff->y_span);
	clSetKernelArg(kernel, 6, sizeof(unsigned int), (void *)&buff->width);
	clSetKernelArg(kernel, 7, sizeof(unsigned int), (void *)&buff->height);

	ret = clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &data_size, &work_size, 0, NULL, NULL);
		check_error(ret);

	clEnqueueReadBuffer(queue, dat_buff, CL_TRUE, 0, data_size*sizeof(data_array[0]), data_array, 0, NULL, NULL);

	printf("Calculation finished!\n");

	ret = clFlush(queue);
	ret = clFinish(queue);
	ret = clReleaseKernel(kernel);
	ret = clReleaseProgram(program);
	ret = clReleaseMemObject(dat_buff);
	ret = clReleaseCommandQueue(queue);
	ret = clReleaseContext(context);
	free(platform_ids);
	free(device_ids);

	return 0;

}

// This will identify our vertex buffer
GLuint vertexbuffer;

// An array of 3 vectors which represents 3 vertices
static const GLfloat g_vertex_buffer_data[] = {
   -10.0f, -10.0f, 0.0f,
   10.0f, -10.0f, 0.0f,
   0.0f,  10.0f, 0.0f,
};


static gboolean render(GtkGLArea *area, GdkGLContext *context)
{
	gtk_gl_area_make_current(area);
  glClearColor (1, 0, 1, 0.5);
  //glClear (GL_COLOR_BUFFER_BIT);


  // Generate 1 buffer, put the resulting identifier in vertexbuffer
  glGenBuffers(1, &vertexbuffer);
  // The following commands will talk about our 'vertexbuffer' buffer
  glBindBuffer(GL_ARRAY_BUFFER, vertexbuffer);
  // Give our vertices to OpenGL.
  glBufferData(GL_ARRAY_BUFFER, sizeof(g_vertex_buffer_data), g_vertex_buffer_data, GL_STATIC_DRAW);


  // 1st attribute buffer : vertices
  glEnableVertexAttribArray(0);
  glBindBuffer(GL_ARRAY_BUFFER, vertexbuffer);
  glVertexAttribPointer(
     0,                  // attribute 0. No particular reason for 0, but must match the layout in the shader.
     3,                  // size
     GL_FLOAT,           // type
     GL_FALSE,           // normalized?
     0,                  // stride
     (void*)0            // array buffer offset
  );
  // Draw the triangle !
  glDrawArrays(GL_TRIANGLES, 0, 3); // Starting from vertex 0; 3 vertices total -> 1 triangle
  glDisableVertexAttribArray(0);


  return TRUE;
}



int main(int argc, char **argv)
{
	GtkWindow *window;
	GtkBuilder *builder;
	GtkGLArea *glarea;
	static struct canvas_buffer mandelbrot_buff;

	gtk_init(&argc, &argv);

	builder = gtk_builder_new_from_resource("/main.glade");
	glarea = GTK_GL_AREA(gtk_builder_get_object(builder, "glarea"));

	g_object_unref(builder);

	mandelbrot_buff.x_span = 1;//2.8;
	mandelbrot_buff.y_span = 0.5;//2.25;
	mandelbrot_buff.center_x = -0.6;
	mandelbrot_buff.center_y = 0.6;
	mandelbrot_buff.width = 1920*3;
	mandelbrot_buff.height = 1080*3;
	mandelbrot_buff.iterations = 70;
	mandelbrot_buff.mandelbrot_buffer = NULL;
	//calculate_mandelbrot(&mandelbrot_buff);

	gtk_gl_area_make_current(glarea);



	GError *error = NULL;
	g_signal_connect(glarea, "render", G_CALLBACK(render), NULL);


	gtk_main();

	return 0;
}
